/*
**                      Softwarepraktikum iMaze
**                              1993/94
**                Joerg Czeranski    Hans-Ulrich Kiel
**
** Datei: X_daten.c
**
** Kommentar:
**  Verwaltung der Objektlisten
*/


#include <stdio.h>

#include "global.h"
#include "speicher.h"
#include "grafik.h"
#include "X_daten.h"

static char sccsid[] = "@(#)X_daten.c	2.2 12/08/95";


/*
** objekte_kopieren
**  kopiert Liste, in der die Grafikobbjekte fuer die 3D-Darstellung
**  zwischengespeichert werden
**
** Parameter:
**  objekte: Liste, Ziel
**  objekte_neu: Liste, Quelle
**
** Seiteneffekte:
**  objekte_neu wird gesetzt
*/
void objekte_kopieren(objekte, objekte_neu)
struct objektdaten *objekte, *objekte_neu;
{
	int i; /* Index fuer die Kugeln */

	objekte->hintergrund_zeichnen = objekte_neu->hintergrund_zeichnen;

	/* Liste der Waende freigeben und dann neue Kopie erstellen */
	liste_freigeben(&objekte->anzwaende, (void **)&objekte->waende);
	liste_kopieren(&objekte->anzwaende, (void **)&objekte->waende,
		&objekte_neu->anzwaende, (void **)&objekte_neu->waende,
		sizeof(struct flaeche));

	/* in Liste mit Kugeln jeweils Unterliste mit Sichtbarkeitsbereichen
	   freigeben */
	for (i = 0; i < objekte->anzkugeln; i++)
		liste_freigeben(&objekte->kugeln[i].sichtanz,
			(void **)&objekte->kugeln[i].sichtbar);

	/* Liste mit Kugeln freigeben und dann neu kopieren */
	liste_freigeben(&objekte->anzkugeln, (void **)&objekte->kugeln);
	liste_kopieren(&objekte->anzkugeln, (void **)&objekte->kugeln,
		&objekte_neu->anzkugeln, (void **)&objekte_neu->kugeln,
		sizeof(struct kugel));

	/* in Liste mit Kugeln jeweils Unterliste mit Sichtbarkeitsbereichen
	   neu kopieren */
	for (i = 0; i < objekte->anzkugeln; i++)
		liste_kopieren(&objekte->kugeln[i].sichtanz,
			(void **)&objekte->kugeln[i].sichtbar,
			&objekte_neu->kugeln[i].sichtanz,
			(void **)&objekte_neu->kugeln[i].sichtbar,
			sizeof(struct ausschnitt));

	/* alten Spruch des Gegners freigeben */
	if (objekte->text != NULL)
		speicher_freigeben((void **)&objekte->text);

	/* Spruch des Gegners kopieren */
	if (objekte_neu->text == NULL)
		objekte->text = NULL;
	else
	{
		speicher_belegen((void **)&objekte->text,
			strlen(objekte_neu->text) + 1);
		strcpy(objekte->text, objekte_neu->text);
	}
}


/*
** objekt_listen_init
**  initialisiert Liste, in der die Grafikobbjekte fuer die 3D-Darstellung
**  zwischengespeichert werden
**
** Parameter:
**  objekte: Liste, die initialisiert werden soll
**
** Seiteneffekte:
**  *objekte wird gesetzt
*/
void objekt_listen_init(objekte)
struct objektdaten *objekte;
{
	/* zu Anfang ist das Bild leer */
	objekte->hintergrund_zeichnen = 0;

	/* objekte enthaelt ihrerseits eine Liste von Waenden und Kugeln */
	liste_initialisieren(&objekte->anzwaende, (void **)&objekte->waende);
	liste_initialisieren(&objekte->anzkugeln, (void **)&objekte->kugeln);

	/* kein Gegner-Spruch */
	objekte->text = NULL;
}
