/*
**                      Softwarepraktikum iMaze
**                              1993/94
**                Joerg Czeranski    Hans-Ulrich Kiel
**
** Datei: motif_fenster.c
**
** Kommentar:
**  Fensterverwaltung auf Basis von Motif
*/


#include <stdio.h>
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/BulletinB.h>
#include <Xm/CascadeBG.h>
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/TextF.h>

#include "grafik.h"
#include "signale.h"
#include "ereignisse.h"
#include "einaus.h"
#include "client.h"
#include "system.h"
#include "X_grafik.h"
#include "X_daten.h"
#include "X_icon.h"
#include "motif_appdefs.h"

static char sccsid[] = "@(#)motif_fenster.c	2.26 01/09/96";


#ifndef DEFAULT_SERVER
#define DEFAULT_SERVER "imaze.rz.tu-clausthal.de"
#endif


#define STR_LENGTH 256

/* Pfusch! Irix doesn't know XmFONTLIST_DEFAULT_TAG */
#define DEFAULT_FONTLIST_TAG "XmFONTLIST_DEFAULT_TAG_STRING"


struct view
{
	char *name;
	Widget w;
	void *X_data;
	struct fenster window;
	void *data;
	int is_open, is_initialized;
	int was_open;
	void (*init)(), (*redraw)();
};


static XtAppContext app_context;
static Arg icon_arglist[1];

static int toplevel_exists = 0;

static Widget toplevel, status_w;
static Widget connect_button_w, disconnect_button_w;
static Widget pause_button_w, resume_button_w;
static Widget server_field_w, message_field_w;

static struct view frontview;
static struct blickfelddaten frontview_data;

static struct view rearview;
static struct blickfelddaten rearview_data;

static struct view compassview;
static int compass_data;

static struct view scoreview;
static struct punktedaten score_data;

static struct view mapview;
static struct grundrissdaten map_data;

static int connected = 0, in_pause_mode;
static XtInputId network_proc_id;

static int request_result;


static void request_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	request_result = *(int *)client_data;
}


static void view_init(view)
struct view *view;
{
	Arg arglist[3];
	Dimension width, height;
	unsigned int depth;

	XtSetArg(arglist[0], XmNwidth, &width);
	XtSetArg(arglist[1], XmNheight, &height);
	XtSetArg(arglist[2], XmNdepth, &depth);

	XtGetValues(view->w, arglist, 3);

	view->window.breite = width;
	view->window.hoehe = height;
	view->window.farbtiefe = depth;

	view->init(&view->X_data, &view->window);
}


static void view_resize(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	struct view *view;

	view = (struct view *)client_data;
	X_zeichne_sync_anfang(view->window.display);
	X_fenster_freigeben(&view->X_data, &view->window);
	view_init(view);
	X_zeichne_sync_ende(view->window.display);
}


static void view_redraw(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	struct view *view;

	view = (struct view *)client_data;
	X_zeichne_sync_anfang(view->window.display);
	view->redraw(view->X_data, &view->window, view->data);
	X_zeichne_sync_ende(view->window.display);
}


static void view_input(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	X_event_routine(XtDisplay(widget), call_data->event);
}


static Widget create_input_box(name, parent_w)
char *name;
Widget parent_w;
{
	Widget form_w, label_w, text_w;
	Arg arglist[4];

	form_w = XtCreateManagedWidget(name,
		xmFormWidgetClass, parent_w, NULL, 0);

	XtSetArg(arglist[0], XmNtopAttachment, XmATTACH_FORM);
	XtSetArg(arglist[1], XmNbottomAttachment, XmATTACH_FORM);

	label_w = XtCreateManagedWidget("label",
		xmLabelWidgetClass, form_w, arglist, 0);

	XtSetArg(arglist[0], XmNleftAttachment, XmATTACH_WIDGET);
	XtSetArg(arglist[1], XmNleftWidget, label_w);
	XtSetArg(arglist[2], XmNrightAttachment, XmATTACH_FORM);
	XtSetArg(arglist[3], XmNresizable, True);

	text_w = XtCreateManagedWidget("input",
		xmTextFieldWidgetClass, form_w, arglist, 0);

	return text_w;
}


static void view_destroy(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	struct view *view;

	view = (struct view *)client_data;

	view->is_open = 0;
	view->is_initialized = 0;
}


static void popup_view(view)
struct view *view;
{
	if (view->is_open)
		return;

	if (!view->is_initialized)
	{
		view->w = XtCreatePopupShell(view->name,
			topLevelShellWidgetClass, toplevel, icon_arglist,
			sizeof icon_arglist / sizeof *icon_arglist);

		XtAddCallback(view->w, XmNdestroyCallback,
			view_destroy, (XtPointer)view);
	}

	XtPopup(view->w, XtGrabNone);
	view->is_open = 1;

	if (!view->is_initialized)
	{
		Widget draw_area;

		draw_area = XtCreateManagedWidget("draw",
			xmDrawingAreaWidgetClass, view->w, NULL, 0);
		view->window.display = XtDisplay(view->w);
		view->window.window = XtWindow(draw_area);
		view_init(view);

		XtAddCallback(draw_area, XmNexposeCallback,
			view_redraw, (XtPointer)view);
		XtAddCallback(draw_area, XmNresizeCallback,
			view_resize, (XtPointer)view);
		XtAddCallback(draw_area, XmNinputCallback, view_input, NULL);

		view->is_initialized = 1;
	}
}


static void popdown_view(view)
struct view *view;
{
	if (!view->is_open)
		return;

	XtPopdown(view->w);
	view->is_open = 0;
}


static void view_toggle(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	struct view *view;

	view = (struct view *)client_data;

	if (view->is_open)
		popdown_view(view);
	else
		popup_view(view);
}


static void set_pause_mode(in_pause)
{
	Arg arglist[1];

	XtSetArg(arglist[0], XmNsensitive, !in_pause);
	XtSetValues(pause_button_w, arglist, 1);

	XtSetArg(arglist[0], XmNsensitive, in_pause);
	XtSetValues(resume_button_w, arglist, 1);
}


static void do_disconnect()
{
	Arg arglist[1];
	char *server_str;
	char server_name[STR_LENGTH];
	XmString str1, str2, str3;

	if (!connected)
		return;

	connected = 0;

	XtSetArg(arglist[0], XmNsensitive, False);
	XtSetValues(disconnect_button_w, arglist, 1);
	XtSetValues(pause_button_w, arglist, 1);
	XtSetValues(resume_button_w, arglist, 1);

	XtRemoveInput(network_proc_id);

	frontview.was_open = frontview.is_open;
	rearview.was_open = rearview.is_open;
	compassview.was_open = compassview.is_open;
	scoreview.was_open = scoreview.is_open;
	mapview.was_open = mapview.is_open;

	popdown_view(&frontview);
	popdown_view(&rearview);
	popdown_view(&compassview);
	popdown_view(&scoreview);
	popdown_view(&mapview);

	XtSetArg(arglist[0], XmNvalue, &server_str);
	XtGetValues(server_field_w, arglist, 1);

	strcpy(server_name, server_str);

	str1 = XmStringCreate("Disconnecting from ", DEFAULT_FONTLIST_TAG);
	str2 = XmStringCreate(server_name, DEFAULT_FONTLIST_TAG);
	str3 = XmStringConcat(str1, str2);

	XmStringFree(str1);
	XmStringFree(str2);

	XtSetArg(arglist[0], XmNlabelString, str3);
	XtSetValues(status_w, arglist, 1);

	XFlush(XtDisplay(toplevel));

	client_spiel_beenden();

	XtSetArg(arglist[0], XmNlabelString, XmStringCreate("Not connected",
		DEFAULT_FONTLIST_TAG));
	XtSetValues(status_w, arglist, 1);

	XtSetArg(arglist[0], XmNsensitive, True);
	XtSetValues(connect_button_w, arglist, 1);

	XtSetArg(arglist[0], XmNeditable, True);
	XtSetValues(server_field_w, arglist, 1);
	XtSetValues(message_field_w, arglist, 1);
}


static void network_ready_handler(data, source, id)
XtPointer data;
int *source;
XtInputId *id;
{
	if (bewegung(0, 1))
	{
		char **error_message;
		char *button_label;

		milden_fehler_abfragen(&error_message, &button_label);
		ueblen_fehler_anzeigen(error_message, button_label);

		do_disconnect();
	}
}


static void disconnect_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	do_disconnect();
}


static void connect_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	Arg arglist[1];
	char *server_str, *message_str;
	char server_name[STR_LENGTH], message_name[STR_LENGTH];
	XmString str1, str2, str3;
	int *descriptor;

	if (connected)
		return;

	XtSetArg(arglist[0], XmNsensitive, False);
	XtSetValues(connect_button_w, arglist, 1);

	XtSetArg(arglist[0], XmNeditable, False);
	XtSetValues(server_field_w, arglist, 1);
	XtSetValues(message_field_w, arglist, 1);

	XtSetArg(arglist[0], XmNvalue, &server_str);
	XtGetValues(server_field_w, arglist, 1);

	strcpy(server_name, server_str);

	XtSetArg(arglist[0], XmNvalue, &message_str);
	XtGetValues(message_field_w, arglist, 1);

	strcpy(message_name, message_str);

	str1 = XmStringCreate("Connecting to ", DEFAULT_FONTLIST_TAG);
	str2 = XmStringCreate(server_name, DEFAULT_FONTLIST_TAG);
	str3 = XmStringConcat(str1, str2);

	XmStringFree(str1);
	XmStringFree(str2);

	XtSetArg(arglist[0], XmNlabelString, str3);
	XtSetValues(status_w, arglist, 1);

	XFlush(XtDisplay(toplevel));

	if (client_spiel_starten(server_name, message_name))
	{
		char **error_message;
		char *button_label;

		milden_fehler_abfragen(&error_message, &button_label);
		ueblen_fehler_anzeigen(error_message, button_label);

		XtSetArg(arglist[0], XmNsensitive, True);
		XtSetValues(connect_button_w, arglist, 1);

		XtSetArg(arglist[0], XmNeditable, True);
		XtSetValues(server_field_w, arglist, 1);
		XtSetValues(message_field_w, arglist, 1);

		XtSetArg(arglist[0], XmNlabelString,
			XmStringCreate("Not connected", DEFAULT_FONTLIST_TAG));
		XtSetValues(status_w, arglist, 1);

		return;
	}

	if ((descriptor = bewegung_deskriptor()) != NULL)
		network_proc_id = XtAppAddInput(app_context, *descriptor,
			XtInputReadMask, network_ready_handler,
			NULL);

	connected = 1;

	if (frontview.was_open)
		popup_view(&frontview);

	if (scoreview.was_open)
		popup_view(&scoreview);

	if (compassview.was_open)
		popup_view(&compassview);

	if (scoreview.was_open)
		popup_view(&scoreview);

	if (mapview.was_open)
		popup_view(&mapview);

	str1 = XmStringCreate("Connected to ", DEFAULT_FONTLIST_TAG);
	str2 = XmStringCreate(server_name, DEFAULT_FONTLIST_TAG);
	str3 = XmStringConcat(str1, str2);

	XmStringFree(str1);
	XmStringFree(str2);

	XtSetArg(arglist[0], XmNlabelString, str3);
	XtSetValues(status_w, arglist, 1);

	XtSetArg(arglist[0], XmNsensitive, True);
	XtSetValues(disconnect_button_w, arglist, 1);

	set_pause_mode(0);
}


static void pause_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	eingabe_simulieren(STOPSIGNAL);
}


static void resume_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	eingabe_simulieren(WEITERSIGNAL);
}


static void exit_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	do_disconnect();

	XtDestroyWidget(toplevel);
}


static void destroy_handler(widget, client_data, call_data)
Widget widget;
caddr_t client_data;
XmAnyCallbackStruct *call_data;
{
	toplevel_exists = 0;

	exit(0);
}


static void fill_main(main_w)
Widget main_w;
{
	Widget form_w, menubar_w, menu_w;
	Arg arglist[2];
	char *field_value;

	form_w = XtCreateManagedWidget("form",
		xmFormWidgetClass, main_w, NULL, 0);

	menubar_w = XmCreateMenuBar(form_w, "menuBar", arglist, 0);
	XtManageChild(menubar_w);

	menu_w = XmCreatePulldownMenu(menubar_w, "menu", NULL, 0);

	XtAddCallback(XtCreateManagedWidget("exit",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, exit_handler, NULL);

	XtSetArg(arglist[0], XmNsubMenuId, menu_w);
	XtCreateManagedWidget("fileMenu",
		xmCascadeButtonGadgetClass, menubar_w, arglist, 1);

	menu_w = XmCreatePulldownMenu(menubar_w, "menu", NULL, 0);

	XtAddCallback(XtCreateManagedWidget("frontView",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, view_toggle, (XtPointer)&frontview);

	XtAddCallback(XtCreateManagedWidget("rearView",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, view_toggle, (XtPointer)&rearview);

	XtAddCallback(XtCreateManagedWidget("map",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, view_toggle, (XtPointer)&mapview);

	XtAddCallback(XtCreateManagedWidget("compass",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, view_toggle, (XtPointer)&compassview);

	XtAddCallback(XtCreateManagedWidget("score",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, view_toggle, (XtPointer)&scoreview);

	XtSetArg(arglist[0], XmNsubMenuId, menu_w);
	XtCreateManagedWidget("toggleMenu",
		xmCascadeButtonGadgetClass, menubar_w, arglist, 1);

	menu_w = XmCreatePulldownMenu(menubar_w, "menu", NULL, 0);

	XtAddCallback(connect_button_w = XtCreateManagedWidget("connect",
		xmPushButtonGadgetClass, menu_w, NULL, 0),
		XmNactivateCallback, connect_handler, NULL);

	XtSetArg(arglist[0], XmNsensitive, False);
	XtAddCallback(disconnect_button_w = XtCreateManagedWidget("disconnect",
		xmPushButtonGadgetClass, menu_w, arglist, 1),
		XmNactivateCallback, disconnect_handler, NULL);

	XtSetArg(arglist[0], XmNsubMenuId, menu_w);
	XtCreateManagedWidget("serverMenu",
		xmCascadeButtonGadgetClass, menubar_w, arglist, 1);

	menu_w = XmCreatePulldownMenu(menubar_w, "menu", NULL, 0);

	XtSetArg(arglist[0], XmNsensitive, False);
	XtAddCallback(pause_button_w = XtCreateManagedWidget("pause",
		xmPushButtonGadgetClass, menu_w, arglist, 1),
		XmNactivateCallback, pause_handler, NULL);

	XtSetArg(arglist[0], XmNsensitive, False);
	XtAddCallback(resume_button_w = XtCreateManagedWidget("resume",
		xmPushButtonGadgetClass, menu_w, arglist, 1),
		XmNactivateCallback, resume_handler, NULL);

	XtSetArg(arglist[0], XmNsubMenuId, menu_w);
	XtCreateManagedWidget("gameMenu",
		xmCascadeButtonGadgetClass, menubar_w, arglist, 1);

	server_field_w = create_input_box("server", form_w);
	message_field_w = create_input_box("message", form_w);

	XtSetArg(arglist[0], XmNmaxLength, STR_LENGTH - 1);
	XtSetValues(server_field_w, arglist, 1);
	XtSetValues(message_field_w, arglist, 1);

	XtSetArg(arglist[0], XmNvalue, &field_value);
	XtGetValues(server_field_w, arglist, 1);

	if (!field_value[0])
	{
		XtSetArg(arglist[0], XmNvalue, DEFAULT_SERVER);
		XtSetValues(server_field_w, arglist, 1);
	}

	XtSetArg(arglist[0], XmNvalue, &field_value);
	XtGetValues(message_field_w, arglist, 1);

	if (!field_value[0])
	{
		char *message;

		if ((message = benutzer_name()) == NULL)
		{
			speicher_belegen((void **)&message, 10);
			strcpy(message, "Gotcha!");
		}
		else
		{
			speicher_vergroessern((void **)&message,
				strlen(message) + 20);
			strcat(message, " shouts: Gotcha!");
		}

		XtSetArg(arglist[0], XmNvalue, message);
		XtSetValues(message_field_w, arglist, 1);

		speicher_freigeben((void **)&message);
	}

	XtSetArg(arglist[0], XmNlabelString, XmStringCreate("Not connected",
		DEFAULT_FONTLIST_TAG));
	status_w = XtCreateManagedWidget("status",
		xmLabelWidgetClass, form_w, arglist, 1);
}


static int request_window(message, label1, label2, two_buttons, widget_name)
char **message, *label1, *label2, *widget_name;
int two_buttons;
{
	Widget w, q_w, form_w, sep_w, ok_w, cancel_w;
	Widget line_w;
	Arg arglist[3];
	int result_ok = 1, result_cancel = 2;

	w = XmCreateDialogShell(toplevel, "dialog", icon_arglist,
		sizeof icon_arglist / sizeof *icon_arglist);

	q_w = XtCreateWidget(widget_name, xmBulletinBoardWidgetClass,
		w, NULL, 0);

	form_w = XtCreateManagedWidget("form", xmFormWidgetClass,
		q_w, NULL, 0);

	line_w = NULL;
	while (*message != NULL)
	{
		XtSetArg(arglist[0], XmNlabelString,
			XmStringCreate(*message, DEFAULT_FONTLIST_TAG));
		XtSetArg(arglist[1], XmNtopAttachment, XmATTACH_WIDGET);
		XtSetArg(arglist[2], XmNtopWidget, line_w);

		line_w = XtCreateManagedWidget(line_w == NULL ?
			"text1" : "text", xmLabelWidgetClass,
			form_w, arglist, 3);

		message++;
	}

	XtSetArg(arglist[0], XmNtopAttachment, XmATTACH_WIDGET);
	XtSetArg(arglist[1], XmNtopWidget, line_w);
	sep_w = XtCreateManagedWidget("separator", xmSeparatorWidgetClass,
		form_w, arglist, 2);

	ok_w = XtCreateManagedWidget("default", xmPushButtonWidgetClass,
		form_w, NULL, 0);
	if (label1 != NULL)
	{
		XtSetArg(arglist[0], XmNlabelString,
			XmStringCreate(label1, DEFAULT_FONTLIST_TAG));
		XtSetValues(ok_w, arglist, 1);
	}

	if (two_buttons)
	{
		cancel_w = XtCreateManagedWidget("cancel",
			xmPushButtonWidgetClass, form_w, NULL, 0);
		if (label2 != NULL)
		{
			XtSetArg(arglist[0], XmNlabelString,
				XmStringCreate(label2, DEFAULT_FONTLIST_TAG));
			XtSetValues(cancel_w, arglist, 1);
		}

		XtCreateManagedWidget("space", xmPrimitiveWidgetClass,
			form_w, NULL, 0);
	}

	XtSetArg(arglist[0], XmNdefaultButton, ok_w);
	XtSetValues(q_w, arglist, 1);

	if (two_buttons)
	{
		XtSetArg(arglist[0], XmNcancelButton, cancel_w);
		XtSetValues(q_w, arglist, 1);
	}

	XtManageChild(q_w);
	XtRealizeWidget(w);

	if (!two_buttons)
	{
		Dimension form_width, button_width;

		XtSetArg(arglist[0], XmNwidth, &form_width);
		XtGetValues(form_w, arglist, 1);
		XtSetArg(arglist[0], XmNwidth, &button_width);
		XtGetValues(ok_w, arglist, 1);

		XtSetArg(arglist[0], XmNleftOffset,
			((int)form_width - (int)button_width) / 2);
		XtSetValues(ok_w, arglist, 1);
	}

	XtManageChild(w);

	XtAddCallback(ok_w, XmNactivateCallback,
		request_handler, (XtPointer)&result_ok);

	if (two_buttons)
		XtAddCallback(cancel_w, XmNactivateCallback,
			request_handler, (XtPointer)&result_cancel);

	XtAddCallback(w, XmNdestroyCallback, request_handler,
		two_buttons ? (XtPointer)&result_cancel :
			(XtPointer)&result_ok);

	request_result = 0;
	while (!request_result)
	{
		XEvent ev;

		XtAppNextEvent(app_context, &ev);
		XtDispatchEvent(&ev);
	}

	XtRemoveCallback(w, XmNdestroyCallback, request_handler,
		two_buttons ? (XtPointer)&result_cancel :
			(XtPointer)&result_ok);
	XtDestroyWidget(w);
	return request_result;
}


/* bis hier lokaler Teil                       */
/***********************************************/
/* ab hier globaler Teil                       */


void zeichne_sync_anfang()
{
	X_zeichne_sync_anfang(XtDisplay(toplevel));
}


void zeichne_sync_ende()
{
	X_zeichne_sync_ende(XtDisplay(toplevel));
}


void zeichne_blickfeld(objekte_neu)
struct objektdaten *objekte_neu;
{
	struct blickfelddaten data;

	objekte_kopieren(&frontview_data.objekte, objekte_neu);

	if (frontview.is_open)
		X_zeichne_blickfeld(frontview.X_data, &frontview.window,
			&frontview_data);
}


void zeichne_rueckspiegel(objekte_rueck_neu)
struct objektdaten *objekte_rueck_neu;
{
	struct blickfelddaten data;

	objekte_kopieren(&rearview_data.objekte, objekte_rueck_neu);

	if (rearview.is_open)
		X_zeichne_blickfeld(rearview.X_data, &rearview.window,
			&rearview_data);
}


void zeichne_karte(spieler_index, anzkreise_neu, kreise_neu)
int spieler_index, anzkreise_neu;
struct kartenkreis *kreise_neu;
{
	map_data.anzkreise_alt = map_data.anzkreise;
	map_data.kreise_alt = map_data.kreise;
	liste_kopieren(&map_data.anzkreise, (void **)&map_data.kreise,
		&anzkreise_neu, (void **)&kreise_neu,
		sizeof(struct kartenkreis));

	if (mapview.is_open)
		X_zeichne_karte(mapview.X_data, &mapview.window,
			&map_data);

	liste_freigeben(&map_data.anzkreise_alt,
		(void **)&map_data.kreise_alt);
	map_data.anzkreise_alt = 0;
	map_data.kreise_alt = NULL;
}


void zeichne_grundriss(anzlinien_neu, linien_neu)
int anzlinien_neu;
struct kartenlinie *linien_neu;
{
	liste_freigeben(&map_data.anzlinien, (void **)&map_data.linien);
	liste_kopieren(&map_data.anzlinien, (void **)&map_data.linien,
		&anzlinien_neu, (void **)&linien_neu,
		sizeof(struct kartenlinie));

	if (mapview.is_open)
		X_zeichne_grundriss(mapview.X_data, &mapview.window,
			&map_data);
}


void zeichne_kompass(blickrichtung_neu)
int blickrichtung_neu;
{
	compass_data = blickrichtung_neu;

	if (compassview.is_open)
		X_zeichne_kompass(compassview.X_data, &compassview.window,
			&compass_data);
}


void zeichne_punkte(punkte_neu)
struct punktedaten *punkte_neu;
{
	liste_freigeben(&score_data.anzpunkte, (void **)&score_data.punkte);
	liste_kopieren(&score_data.anzpunkte, (void **)&score_data.punkte,
		&punkte_neu->anzpunkte, (void **)&punkte_neu->punkte,
		sizeof(struct punktestand));

	if (scoreview.is_open)
		X_zeichne_punkte(scoreview.X_data, &scoreview.window,
			&score_data);
}


void grafik_schleife()
{
	connected = 0;

	XtAppMainLoop(app_context);
}


int grafik_init(argc, argv)
int *argc;
char **argv;
{
	toplevel = XtAppInitialize(&app_context, "IMaze",
		NULL, 0, argc, argv, fallback_app_ressources, NULL, 0);
	XtAddCallback(toplevel, XmNdestroyCallback, destroy_handler, NULL);

	toplevel_exists = 1;

	fill_main(toplevel);
	XtRealizeWidget(toplevel);

	XtSetArg(icon_arglist[0], XmNiconPixmap,
		X_create_icon_pixmap(XtDisplay(toplevel),
		XScreenNumberOfScreen(XtScreen(toplevel))));

	XtSetValues(toplevel, icon_arglist, 1);

	if (X_farben_init(XtDisplay(toplevel),
		XScreenNumberOfScreen(XtScreen(toplevel))))
	{
		XtRemoveCallback(toplevel, XmNdestroyCallback,
			destroy_handler, NULL);
		XtDestroyWidget(toplevel);

		toplevel_exists = 0;

		return 1;
	}

	frontview.name = "frontView";
	frontview.data = &frontview_data;
	frontview.is_open = 0;
	frontview.is_initialized = 0;
	frontview.was_open = 1;
	frontview.init = X_fenster_init;
	frontview.redraw = X_zeichne_blickfeld;
	objekt_listen_init(&frontview_data.objekte);
	frontview_data.fadenkreuz = 1;

	scoreview.name = "score";
	scoreview.data = &score_data;
	scoreview.is_open = 0;
	scoreview.is_initialized = 0;
	scoreview.was_open = 1;
	scoreview.init = X_fenster_init;
	scoreview.redraw = X_zeichne_punkte;
	liste_initialisieren(&score_data.anzpunkte,
		(void **)&score_data.punkte);

	compassview.name = "compass";
	compassview.data = &compass_data;
	compassview.is_open = 0;
	compassview.is_initialized = 0;
	compassview.was_open = 0;
	compassview.init = X_fenster_init;
	compassview.redraw = X_zeichne_kompass;
	compass_data = -1;

	rearview.name = "rearView";
	rearview.data = &rearview_data;
	rearview.is_open = 0;
	rearview.is_initialized = 0;
	rearview.was_open = 0;
	rearview.init = X_fenster_init;
	rearview.redraw = X_zeichne_blickfeld;
	objekt_listen_init(&rearview_data.objekte);
	rearview_data.fadenkreuz = 0;

	mapview.name = "map";
	mapview.data = &map_data;
	mapview.is_open = 0;
	mapview.is_initialized = 0;
	mapview.was_open = 0;
	mapview.init = X_karte_init;
	mapview.redraw = X_zeichne_grundriss;
	liste_initialisieren(&map_data.anzlinien, (void **)&map_data.linien);
	liste_initialisieren(&map_data.anzkreise, (void **)&map_data.kreise);
	map_data.anzkreise_alt = 0;
	map_data.kreise_alt = NULL;

	if (X_eingabe_init(argc, argv))
		return 1;

	return 0;
}


void grafik_ende()
{
	if (toplevel_exists)
		XtDestroyWidget(toplevel);

	X_eingabe_ende();
}


void ueblen_fehler_anzeigen(meldung, knopf)
char **meldung, *knopf;
{
	if (!toplevel_exists)
	{
		while (*meldung != NULL)
			fprintf(stderr, "iMaze: %s\n", *meldung++);

		return;
	}

	if (connected)
	{
		connected = 0;
		XtRemoveInput(network_proc_id);
	}

	request_window(meldung, knopf, NULL, 0, "error");
}


void ereignisse_darstellen(ereignisse)
char ereignisse[ERG_ANZ];
{
	if (ereignisse[ABGESCHOSSEN_ERG] || ereignisse[GETROFFEN_ERG] ||
		ereignisse[AKTION_FEHLER_ERG] || ereignisse[SUSPENDIERT_ERG])
		XBell(XtDisplay(toplevel), 0);

	if (ereignisse[PAUSE_ERG])
		set_pause_mode(1);
	else if (ereignisse[ERWACHT_ERG])
		set_pause_mode(0);
}


int rueckfrage(meldung, weiter_knopf, abbruch_knopf)
char **meldung, *weiter_knopf, *abbruch_knopf;
{
	if (!toplevel_exists)
	{
		static char *message[] = { "iMaze - Fatal Error", "",
			"Couldn't open notice window.", NULL };

		uebler_fehler(message, NULL);
	}

	return request_window(meldung, weiter_knopf,
		abbruch_knopf, 1, "question") == 1;
}


void disconnect_abfrage()
{
	static char *message[] = { "Do you really want do disconnect?", NULL };

	if (!rueckfrage(message, "Sure", "Play on"))
		return;

	do_disconnect();
}
