/*
**                      Softwarepraktikum iMaze
**                              1993/94
**                Joerg Czeranski    Hans-Ulrich Kiel
**
** Datei: xv_tasten.c
**
** Kommentar:
**  Wertet Events von der Tastatur aus und setzt definierte Signale
*/


#include <xview/xview.h>
#include <xview/canvas.h>
#include <xview/panel.h>

#include "global.h"
#include "signale.h"
#include "ereignisse.h"
#include "einaus.h"
#include "grafik.h"
#include "xv_einaus.h"

static char sccsid[] = "@(#)xv_tasten.c	2.1 12/08/95";


/*
** event_routine
**  verarbeitet die Tastatur-Events und ordnet sie den internen Signalen zu
**
** Parameter:
**  window: Fenster fuer den der Event bestimmt ist
**  event: Beschreibung des Events
**
** Seiteneffekte:
**  aktiv wird veraendert
*/
static void event_routine(window, event)
Xv_Window window;
Event *event;
{
	X_event_routine(XV_DISPLAY_FROM_WINDOW(event_window(event)),
		event_xevent(event));
}


/*
** frame_init
**  sucht in einem Fenster nach Unterfenstern vom Typ CANVAS und
**  initialisiert darin die Event-Abfrage
**
** Parameter:
**  frame: Fenster, in dem gesucht werden soll
*/
static void frame_init(frame)
Frame frame;
{
	int i;
	Window subwindow;

	for (i = 1; (subwindow = xv_get(frame,
		FRAME_NTH_SUBWINDOW, i)) != 0; i++)
		if ((Xv_pkg *)xv_get(subwindow, XV_TYPE) == CANVAS)
		{
			/* KBD_DONE: Fenster verliert Fokus
			   WIN_ASCII_EVENT: Tasten werden gedrueckt
			   WIN_UP_EVENT: Taste losgelassen (wird evtl.
			   nicht unterstuetzt) */
			xv_set(canvas_paint_window(subwindow),
				WIN_CONSUME_EVENTS,
					KBD_DONE, WIN_ASCII_EVENTS,
					WIN_UP_EVENTS, 0,
				WIN_EVENT_PROC, event_routine,
				0);
		}
}


/* bis hier lokaler Teil                       */
/***********************************************/
/* ab hier globaler Teil                       */


/*
** xv_eingabe_init
**  sucht nach Unterfenstern vom Typ CANVAS und initialisiert darin
**  die Event-Abfrage
**
** Parameter:
**  hauptframe: Fenster, von dem alle anderen Fenster abstammen
**  panel: Panel, in dem Buttons ergaenzt werden duerfen
**
** Rueckgabewert:
**  0 fuer Erfolg, 1 fuer Fehler
*/
int xv_eingabe_init(hauptframe, panel, argc, argv)
Frame hauptframe;
Panel panel;
int *argc;
char **argv;
{
	int i;          /* Index */
	Frame subframe; /* ein Fenster */

	/* im Hauptfenster suchen */
	frame_init(hauptframe);

	/* in allen weiteren Fenstern suchen */
	for (i = 1; (subframe = xv_get(hauptframe,
		FRAME_NTH_SUBFRAME, i)) != 0; i++)
		frame_init(subframe);

	return X_eingabe_init(argc, argv);
}


/*
** xv_eingabe_ende
**  Dummy-Routine fuer das Schliessen von Eingabekanaelen etc.
*/
void xv_eingabe_ende()
{
	X_eingabe_ende();
}
